/*
 *  rommon Simple Monitor
 * 
 *  Copyright (C) 1999-2004 by Dep. of Computer Science and Engineering
 *                   Tomakomai National College of Technology, JAPAN
 *  Copyright (C) 2006-2015 by GJ Business Division
 *                   RICOH COMPANY,LTD. JAPAN
 *  Copyright (C) 2016-2019 by TOPPERS PROJECT Educational Working Group.
 *
 *  嵭Ԥϡʲ (1)(4) ξ狼Free Software Foundation 
 *  ˤäƸɽƤ GNU General Public License  Version 2 ˵
 *  ҤƤ˸¤ꡤܥեȥܥեȥ
 *  ѤΤޤࡥʲƱˤѡʣѡۡʰʲ
 *  ѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơŬѲǽ
 *  ޤơʤݾڤԤʤޤܥեȥѤˤľ
 *  ŪޤϴŪʤ»˴ؤƤ⡤Ǥʤ
 * 
 *  @(#) $Id: $
 */

/*
 *	monmain.c -- rom Monitor
 */

#include <stdio.h>
#include "comdefs.h"
#include "mondefs.h"
#include "serial.h"

extern void (*BeforeInitEntry)(void);
extern void (*AfterInitEntry)(void);

static void InitTerm(void);
static void InitVect(void);
static void LoadCommand(int id, char **v);
static void GoCommand(int id, char **ps);
static void HelpCmd(int id, char **ps);
#ifdef USE_VOLUME
static void FlCommand(int id, char **ps);
static void LsCommand(int id, char **ps);
#endif

/*
 *	Command String and Vector
 */
const long CommandVector[] = {
	(long)EnterLongCmd,			/* enter long data command */
	(long)EnterWordCmd,			/* enter word data command */
	(long)EnterByteCmd,			/* enter byte data command */
	(long)DumpLongCmd,			/* dump long command */
	(long)DumpWordCmd,			/* dump word command */
	(long)DumpByteCmd,			/* dump byte command */
	(long)FillByteCmd,			/* fill byte command */
	(long)LoadCommand,			/* load command */
	(long)GoCommand,			/* go command */
#ifdef USE_VOLUME
	(long)FlCommand,			/* fl command */
	(long)LsCommand,			/* ls commnad */
#endif
	(long)HelpCmd
};

const unsigned short CommandString[] = {
	MONCMD_ENTER_LONG,
	MONCMD_ENTER_WORD,
	MONCMD_ENTER_BYTE,
	MONCMD_DUMP_LONG,
	MONCMD_DUMP_WORD,
	MONCMD_DUMP_BYTE,
	MONCMD_FILL_BYTE,
	MONCMD_LOAD,
	MONCMD_GO,
#ifdef USE_VOLUME
	MONCMD_FL,
	MONCMD_LS,
#endif
	CMDSTR_HELP,
	0			/* terminator */
};

static const char HelpStrings[] = 
		TARGET_CPU_STR
		" Monitor v1.03\r\n"
		"Copyright (C) 1999-2004 CSE Tomakomai NCT\r\n"
		"Copyright (C) 2016-2019 TOPPERS PROJECT.\0"
		"----------------------------------------------------------------------------\0"
		"ld             load program          | go[<ad>]       go program\0"
		"db[<ad>[<ln>]] dump byte             | dw[<ad>[<ln>]] dump word\0"
		"dl[<ad>[<ln>]] dump long word        | eb[<ad>]       enter byte\0"
		"ew[<ad>[<ln>]] enter word            | el[<ad>]       enter long word\0"
#ifdef USE_VOLUME
		"fl[<fn><ad><dz>] load file           | ls[<ph>]       display dir \0"
#endif
		"fb[<ad>[<ln>[<in>[<ic>[<rp>]]]]] fill byte\0"
		"?              help\0"
		"\0";

/*
 *	Start Monitor
 */
void MonMain(void)
{
	SetVector((void *)MonitorVector);
	if(BeforeInitEntry)
		BeforeInitEntry();
	InitVect();
	InitTerm();
	if(AfterInitEntry)
		AfterInitEntry();
	Enable();
#if USE_VOLUME
	VolumeInit();
#endif
#ifdef MULTIPLE_CONSOLE
	for(id = 0 ; id < MAX_SERIAL ; id++){
		_setup_stdio(id);
		printf("%s\n", HelpStrings);
		SendPrompt(id);
	}
	for(;;){
		for(id = 0 ; id < MAX_SERIAL ; id++){
			_setup_stdio(id);
			ParseCmdCore(id);
		}
	}

#else	/* of #ifdef MULTIPLE_CONSOLE */
#ifdef LOAD_AND_GO_ONLY
	_setup_stdio(DEF_MON_SERIAL);
	printf("%s\n", HelpStrings);
	InitRecLine(DEF_MON_SERIAL);
	LoadCommand(DEF_MON_SERIAL, 0);
	InputLine[DEF_MON_SERIAL][0] = '\0';
	GoCommand(DEF_MON_SERIAL);
	for(;;){}

#else	/* of #ifdef LOAD_AND_GO_ONLY */
	_setup_stdio(DEF_MON_SERIAL);
	printf("%s\n", HelpStrings);
	SendPrompt(DEF_MON_SERIAL);
	for(;;){
		ParseCmdCore(DEF_MON_SERIAL);
	}

#endif	/* of #ifdef LOAD_AND_GO_ONLY */

#endif	/* of #ifdef MULTIPLE_CONSOLE */
}

/*	
 *	Initialize Terminal
 */
static void InitTerm(void)
{
	int id;

	for(id = 0 ; id < MAX_SERIAL ; id++){
		InitSerial(id);							/* Initialize serial port */
		AddressPointer[id] = (char *)RAM_BASE;	/* Initialize address pointer */
		DumpPointer[id] = (char *)RAM_BASE;		/* Initialize dump    pointer */
		CommandMode[id] = CMDMODE_ADDR;			/* Initialize command mode */
	}
}

/*
 *	Initialize Intrrupt Vectors
 */
static void InitVect(void)
{
#ifndef NOT_INITDECT
	MoveLongInc(UserVector, MonitorVector, VECTOR_SIZE);
#endif
}

/*
 *	Send Prompt	
 *	Input:	arg1: Serial Channel #, 0 or 1
 */
void SendPrompt(int id)
{
	MonSendPrompt(id);
	Status[id] = 0;
	InitRecLine(id);
}

/*
 *	Check Sum
 *	Input:	er3: Input Line + 1
 *	Output:	r0l: Check Sum, 0 if no error
 */
static unsigned char CheckSum(char *s)
{
	int len;
	unsigned int Sum = 0;

	len = HexASCII2Long(&s, 2);
	s -= 2;
	do{
		Sum += HexASCII2Long(&s, 2);
		len--;
	}while(len >= 0);
	return Sum;
}

/*
 *	Load Command
 *	Format: ld
 *	Input:  arg1: Serial Channel #, 0 or 1
 *          arg2: input line
 */
static void LoadCommand(int id, char **v)
{
	unsigned long e = 0;
	char *ea = 0;
	char *sa = 0;
	char *a  = 0;
	char *s;
	char c;
	int  len, alen;
	unsigned char Sum, val;
	int  cnt=0;

	CommandMode[id] = CMDMODE_ADDR;
	do{
		do{
			InitRecLine(id);
			do{
				s = ReciveLine(DUMMY_MON_SERIAL, id);
			}while(s == 0);
		}while(*s++ != 'S');
		Sum = CheckSum(s+1);
		if(Sum != 0x00 && Sum != 0xff){
			cnt++;
			ea = a;
			if(Status[id] == 0)
				Status[id] = ERR_CHECK_SUM;	/* check sum error */
		}
		c = *s++;
		len = HexASCII2Long(&s, 2)-1;
		if(c == '1')
			alen = 4;
		else if(c == '2')
			alen = 6;
		else if(c == '3')
			alen = 8;
		else if(c == '7'){
			e = (unsigned long)HexASCII2Long(&s, 4);
			alen = 0;
		}
		else if(c == '8'){
			e = (unsigned long)HexASCII2Long(&s, 6);
			alen = 0;
		}
		else if(c == '9'){
			e = (unsigned long)HexASCII2Long(&s, 8);
			alen = 0;
		}
		else
			alen = 0;
		if(alen > 0){
			a = (char *)HexASCII2Long(&s, alen);
			if(sa == 0)
				sa = a;
			len -= alen/2;
			while(len > 0){
				val = HexASCII2Long(&s, 2);
				*a++ = val;
				len--;
			}
		}
	}while(c < '7' || c > '9');

	printf("Start Address = %08x\n", (int)sa);	/* start address */
	printf("End   Address = %08x\n", (int)a);	/* end address */
	printf("Entry Address = %08x\n", (int)e);
	if(Status[id] != 0)			/* error */
		printf("Error = %08x\n", (int)ea);
}

/*
 *	Go Command
 *	Format: go [<addr>]
 *	Input:  arg1: Serial Channel #, 0 or 1
 *          arg2: input line
 */
static void GoCommand(int id, char **ps)
{
	char *s;
	void (*func)(void);

	s = *ps;
	func = (void (*)(void))HexASCII2Long(ps, 128);
#if defined(NOT_DEFVECTORTYPE)
	if(s == *ps)
		func = (void (*)(void))UserVector;
#elif !defined(RESET_VECTOR)
	if(s == *ps)
		func = (void (*)(void))UserVector[0];
#else
	if(s == *ps)
		func = (void (*)(void))UserVector[RESET_VECTOR];
#endif
	execDisable();
	Disable();
	SetVector((void *)UserVector);
	func();
	Disable();
	SetVector((void *)MonitorVector);
	Enable();
}

#ifdef USE_VOLUME
/*
 *	Fl Command
 *	Format: fl [<name> <addr> <comp size>]
 *	Input:  arg1: Serial Channel #, 0 or 1
 *          arg2: input line
 */
static void FlCommand(int id, char **ps)
{
	char *name, *s;
	unsigned long addr, dsize;

	addr = 0;
	dsize = 0;
	name = *ps;
	if(name[0] == 0){
		printf("no input file !\n");
		return;
	}
	s = *ps;
	while(*s != ' ' && *s != 0){
		s++;
	}
	if(*s == ' ')
	  *s++ = 0;
	*ps = s;
	if(*s != 0){
		SkipBlanks(ps);
		addr = (unsigned long)HexASCII2Long(ps, 128);
		SkipBlanks(ps);
		dsize = (unsigned long)HexASCII2Long(ps, 128);
	}
	load_file(id, addr, dsize, name);
}

/*
 *	Ls Command
 *	Format: ls [<path>]
 *	Input:  arg1: Serial Channel #, 0 or 1
 *          arg2: input line
 */
static void LsCommand(int id, char **ps)
{
	disp_dir(id, *ps);
}
#endif

/*
 *	Help Command
 *	Command Format: ?
 *	Input:	arg1: Serial Channel #, 0 or 1
 *          arg2: input line
 */
static void HelpCmd(int id, char **ps)
{
	char *s = (char *)HelpStrings;

	do{
		do{
			putchar(*s++);
		}while(*s != 0);
		s++;
		putchar('\n');
	}while(*s != 0);
}


