/*
 *  rommon Simple Monitor
 * 
 *  Copyright (C) 1999-2004 by Dep. of Computer Science and Engineering
 *                   Tomakomai National College of Technology, JAPAN
 *  Copyright (C) 2006-2015 by GJ Business Division
 *                   RICOH COMPANY,LTD. JAPAN
 *  Copyright (C) 2016-2019 by TOPPERS PROJECT Educational Working Group.
 *
 *  L쌠҂́Cȉ (1)`(4) ̏CFree Software Foundation 
 *  ɂČ\Ă GNU General Public License  Version 2 ɋL
 *  qĂ𖞂ꍇɌC{\tgEFAi{\tgEFA
 *  ς̂܂ށDȉjgpEEρEĔzziȉC
 *  pƌĂԁj邱Ƃ𖳏ŋD
 *  (1) {\tgEFA\[XR[ȟ`ŗpꍇɂ́CL̒
 *      \C̗pщL̖ۏ؋K肪Ĉ܂܂̌`Ń\[
 *      XR[hɊ܂܂Ă邱ƁD
 *  (2) {\tgEFACCu`ȂǁC̃\tgEFAJɎg
 *      pł`ōĔzzꍇɂ́CĔzzɔhLgip
 *      ҃}jAȂǁjɁCL̒쌠\C̗pщL
 *      ̖ۏ؋Kfڂ邱ƁD
 *  (3) {\tgEFAC@ɑgݍނȂǁC̃\tgEFAJɎg
 *      płȂ`ōĔzzꍇɂ́Ĉꂩ̏𖞂
 *      ƁD
 *    (a) ĔzzɔhLgip҃}jAȂǁjɁCL̒
 *        쌠\C̗pщL̖ۏ؋Kfڂ邱ƁD
 *    (b) Ĕzž`ԂCʂɒ߂@ɂāCTOPPERSvWFNg
 *        񍐂邱ƁD
 *  (4) {\tgEFA̗pɂ蒼ړI܂͊ԐړIɐ邢Ȃ鑹
 *      QCL쌠҂TOPPERSvWFNgƐӂ邱ƁD
 *
 *  {\tgEFÁCۏ؂Œ񋟂Ă̂łDL쌠҂
 *  TOPPERSvWFNǵC{\tgEFAɊւāC̓Kp\
 *  ܂߂āCȂۏ؂sȂD܂C{\tgEFA̗pɂ蒼
 *  ړI܂͊ԐړIɐȂ鑹QɊւĂC̐ӔC𕉂ȂD
 * 
 *  @(#) $Id: $
 */

#include "sys_defs.h"
#include <sil.h>
#include "comdefs.h"
#include "serial.h"

/*
 *  SIL֐̃}N`
 */
#define sil_orw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) | (b))
#define sil_andw_mem(a, b)		sil_wrw_mem((a), sil_rew_mem(a) & ~(b))
#define sil_modw_mem(a, b, c)	sil_wrw_mem((a), (sil_rew_mem(a) & (~b)) | (c))

typedef struct sio_port_control_block {
	uint32_t      base;
    uint32_t      iof;
} sio_port_control_block_t;


static const sio_port_control_block_t sio_pcb_table[2] = {
	{TADR_UART0_BASE, IOF0_UART0_MASK},
	{TADR_UART1_BASE, IOF0_UART1_MASK}
};


/*
 *  none-interrupt UART functions.
 */
static void sys_putc(char c)
{
	uint32_t base = sio_pcb_table[0].base;
	while ((int32_t)sil_rew_mem((uint32_t *)(base+TOFF_UART_TXFIFO)) < 0);
	sil_wrw_mem((uint32_t *)(base+TOFF_UART_TXFIFO), c);
	if(c == '\n'){
		while ((int32_t)sil_rew_mem((uint32_t *)(base+TOFF_UART_TXFIFO)) < 0);
		sil_wrw_mem((uint32_t *)(base+TOFF_UART_TXFIFO), '\r');
	}
}

void sys_printf(const char * s) {
	while (*s != '\0'){
		sys_putc(*s++);
	}
}

void put_hex(char a, int val)
{
	int i, j;
	sys_putc(a);
	sys_putc(' ');
	for(i = 28 ; i >= 0 ; i-= 4){
		j = (val >> i) & 0xf;
		if(j > 9)
			j += ('A'-10);
		else
			j += '0';
		sys_putc(j);
	}
	sys_putc('\n');
}

/*
 *  UART0 Interrupt Handler
 */
void uart_0_handler(void)
{
	uint32_t base = sio_pcb_table[0].base;
	uint32_t status = sil_rew_mem((uint32_t *)(base+TOFF_UART_IP));
	if((status & UART_IP_RXWM) != 0)
		ComInterRX(0);
#ifndef NOINT_UARTTX
	if((status & UART_IP_TXWM) != 0){
		while((int32_t)sil_rew_mem((uint32_t *)(base+TOFF_UART_TXFIFO)) >= 0){
			ComInterTX(0);
			if((sil_rew_mem((uint32_t *)(base+TOFF_UART_IE)) & UART_IP_TXWM) == 0)
				break;
		}
	}
#endif
};


void execDisable(void)
{
	clear_csr(mie, MIP_MEIP);
}

/*
 *  Get RX Data
 *  Input: id: Serial Channel #
 *  output: c: Rx Data
 */
unsigned short GetRxData(int id)
{
	uint32_t base = sio_pcb_table[id].base;
	return sil_rew_mem((uint32_t *)(base+TOFF_UART_RXFIFO)) | CHKRXDATA;
}

/*
 *  Enable Tx interrupt
 *  Input: id: Serial Channel #
 */
void EnableITX(int id)
{
	uint32_t base = sio_pcb_table[id].base;
	sil_orw_mem((uint32_t *)(base+TOFF_UART_IE), UART_IP_TXWM);
#ifndef NOINT_UARTTX
	while((int32_t)sil_rew_mem((uint32_t *)(base+TOFF_UART_TXFIFO)) >= 0){
		ComInterTX(0);
		if((sil_rew_mem((uint32_t *)(base+TOFF_UART_IE)) & UART_IP_TXWM) == 0)
			break;
	}
#endif	/* ROI DEBUG */
}

/*
 *  Disable Tx interrupt
 *  Input: id: Serial Channel #
 */
void DisableITX(int id)
{
	uint32_t base = sio_pcb_table[id].base;
	sil_andw_mem((uint32_t *)(base+TOFF_UART_IE), UART_IP_TXWM);
}

/*
 *  Send Tx in interrupt
 *  Input: id: Serial Channel #
 *         c : Send charactor
 */
void SendTX(int id, char c)
{
	uint32_t base = sio_pcb_table[id].base;
	sil_wrw_mem((uint32_t *)(base+TOFF_UART_TXFIFO), (c & 0xFF));
}

/*
 * Send Tx wait
 * Inputs: id: Serial Channel #
 */
void SendTXWait(int id)
{
	uint32_t base = sio_pcb_table[id].base;
	while((int32_t)sil_rew_mem((uint32_t *)(base+TOFF_UART_TXFIFO)) < 0);
}

void InitSIO(int id)
{
	const sio_port_control_block_t *sio = &sio_pcb_table[id];

	/*
	 *  Congfigurate UART0 GPIO
	 */
	sil_orw_mem((uint32_t *)(TADR_GPIO_BASE+TOFF_GPIO_OUTPUT_VAL), sio->iof);
	sil_orw_mem((uint32_t *)(TADR_GPIO_BASE+TOFF_GPIO_OUTPUT_EN), sio->iof);
	sil_andw_mem((uint32_t *)(TADR_GPIO_BASE+TOFF_GPIO_IOF_SEL), sio->iof);
	sil_orw_mem((uint32_t *)(TADR_GPIO_BASE+TOFF_GPIO_IOF_EN), sio->iof);


	/*
	 *  Configurate UART0 DEFAULT_SPEED Baud Rate
	 */
	sil_wrw_mem((uint32_t *)(sio->base+TOFF_UART_DIV), cpu_frequency / DEFAULT_SPEED - 1);
	sil_wrw_mem((uint32_t *)(sio->base+TOFF_UART_TXCTRL), UART_TXEN + 0x10000);
	sil_wrw_mem((uint32_t *)(sio->base+TOFF_UART_RXCTRL), UART_RXEN);

	/*
	 *  Wait sometimes.
	 */
	sil_dly_nse(10000);

	/*
	 * Configurate Interrupt
	 */
	clear_csr(mie, MIP_MEIP);
	clear_csr(mie, MIP_MTIP);

	plic_init();

	plic_set_vector(IRQ_VECTOR_UART0, uart_0_handler);
	plic_enable_interrupt(IRQ_VECTOR_UART0);
	plic_set_priority(IRQ_VECTOR_UART0, 1);
	sil_wrw_mem((uint32_t *)(sio->base+TOFF_UART_IE), UART_IP_RXWM);

	/*
	 *  Enable the Machine-Interrupt
	 */
	set_csr(mie, MIP_MEIP);

	/*
	 *  Enable PLIC-Interrupt
	 */
	Enable();
}

