/*
 *  rommon Simple Monitor
 * 
 *  Copyright (C) 1999-2004 by Dep. of Computer Science and Engineering
 *                   Tomakomai National College of Technology, JAPAN
 *  Copyright (C) 2015-2019 by Roi Takeuchi
 *
 *  嵭Ԥϡʲ (1)(4) ξ狼Free Software Foundation 
 *  ˤäƸɽƤ GNU General Public License  Version 2 ˵
 *  ҤƤ˸¤ꡤܥեȥܥեȥ
 *  ѤΤޤࡥʲƱˤѡʣѡۡʰʲ
 *  ѤȸƤ֡ˤ뤳Ȥ̵ǵ롥
 *  (1) ܥեȥ򥽡ɤηѤˤϡ嵭
 *      ɽѾ浪Ӳ̵ݾڵ꤬Τޤޤηǥ
 *      ˴ޤޤƤ뤳ȡ
 *  (2) ܥեȥ򡤥饤֥ʤɡ¾Υեȥȯ˻
 *      ѤǤǺۤˤϡۤȼɥȡ
 *      ԥޥ˥奢ʤɡˤˡ嵭ɽѾ浪Ӳ
 *      ̵ݾڵǺܤ뤳ȡ
 *  (3) ܥեȥ򡤵Ȥ߹ʤɡ¾Υեȥȯ˻
 *      ѤǤʤǺۤˤϡΤ줫ξ
 *      ȡ
 *    (a) ۤȼɥȡѼԥޥ˥奢ʤɡˤˡ嵭
 *        ɽѾ浪Ӳ̵ݾڵǺܤ뤳ȡ
 *    (b) ۤη֤̤ˡˤäơTOPPERSץȤ
 *        𤹤뤳ȡ
 *  (4) ܥեȥѤˤľŪޤϴŪ뤤ʤ»
 *      ⡤嵭ԤTOPPERSץȤդ뤳ȡ
 *
 *  ܥեȥϡ̵ݾڤ󶡤ƤΤǤ롥嵭Ԥ
 *  TOPPERSץȤϡܥեȥ˴ؤơŬѲǽ
 *  ޤơʤݾڤԤʤޤܥեȥѤˤľ
 *  ŪޤϴŪʤ»˴ؤƤ⡤Ǥʤ
 * 
 *  @(#) $Id: $
 */

/*
 *	rommon Config Functions
 */
#include "sys_defs.h"
#include <string.h>
#include "sil.h"

/*
 *  ץstm32g0xx-nucleo64ѡ
 */

#define sil_modw_mem(addr, mask, val)	sil_wrw_mem((addr), ((sil_rew_mem(addr) & ~(mask)) | (val)))
#define sil_andw_mem(addr, mask)			sil_wrw_mem((addr), (sil_rew_mem(addr) & ~(mask)))
#define sil_orw_mem(addr, val)			sil_wrw_mem((addr), (sil_rew_mem(addr) | (val)))


#define HSI_TIMEOUT_VALUE          (2000U)      /* 2 ms (minimum Tick + 1U) */
#define PLL_TIMEOUT_VALUE          (2000U)      /* 2 ms (minimum Tick + 1U) */
#define MSI_TIMEOUT_VALUE          (2000U)      /* 2 ms (minimum Tick + 1U) */
#define RCC_DBP_TIMEOUT_VALUE      (2000U)		/* 2 ms (minimum Tick + 1)  */
#define RCC_LSE_TIMEOUT_VALUE      (5000000U)	/* 5 s    */
#define CLOCKSWITCH_TIMEOUT_VALUE  (5000000U)	/* 5 s    */

#if !defined(HSE_VALUE)
#define HSE_VALUE                  8000000  /* Value of the External oscillator in Hz */
#endif /* HSE_VALUE */
#if !defined(HSI_VALUE)
#define HSI_VALUE                  16000000	/* Value of the Internal oscillator in Hz*/
#endif /* HSI_VALUE */
#if !defined  (LSI_VALUE)
#define LSI_VALUE                  32000	/* Value of the Internal Low Speed oscillator in Hz */
#endif /* LSI_VALUE */
#if !defined(LSE_VALUE)
#define LSE_VALUE                  32768	/* Value of the External low oscillator in Hz*/
#endif

/* PLL_VCO = (HSE_VALUE or HSI_VALUE / PLL_M) * PLL_N */
#define PLL_M                      4		/* Possible value 0 and 63 */
#define PLL_N                      70		/* 80 per 64MHz */

/* SYSCLK = PLL_VCO / PLL_P */
#define PLL_P                      10

#define PLL_Q                      5
#define PLL_R                      5	/* Possible value between 2 and 7 */


#define PLLM_SHIFT                 4
#define PLLN_SHIFT                 8
#define PLLR_SHIFT                 29
#define HSIDIV_SHIFT               11
#define ICSCR_HSITRIM_SHIFT        8
#define CFGR_SWS_SHIFT             3
#define CFGR_HPRE_SHIFT            8

#define RCC_PLLSOURCE_NONE         0x00000000				/* No clock selected as PLL entry clock source  */
#define RCC_PLLSOURCE_HSI          RCC_PLLCFGR_PLLSRC_HSI	/* HSI clock selected as PLL entry clock source */
#define RCC_PLLSOURCE_HSE          RCC_PLLCFGR_PLLSRC_HSE	/* HSE clock selected as PLL entry clock source */

#define RCC_HSICALIBRATION_DEFAULT 64						/* Default HSI calibration trimming value */


/*
 *  FLASH_Latency
 */ 
#define FLASH_LATENCY_0            0x00000000			/* FLASH Zero wait state */
#define FLASH_LATENCY_1            FLASH_ACR_LATENCY_0	/* FLASH One wait state */
#define FLASH_LATENCY_2            FLASH_ACR_LATENCY_1	/* FLASH Two wait states */

/*
 *  RCC System Clock Source
 */
#define RCC_SYSCLKSOURCE_HSI       0x00000000						/* HSI selection as system clock */
#define RCC_SYSCLKSOURCE_HSE       RCC_CFGR_SW_0					/* HSE selection as system clock */
#define RCC_SYSCLKSOURCE_PLLCLK    RCC_CFGR_SW_1					/* PLL selection as system clock */
#define RCC_SYSCLKSOURCE_LSI       (RCC_CFGR_SW_1 | RCC_CFGR_SW_0)	/* LSI selection as system clock */
#define RCC_SYSCLKSOURCE_LSE       RCC_CFGR_SW_2					/* LSE selection as system clock */

/*
 *  Div HSI Divided
 */
#define RCC_HSI_DIV1               0x00000000						/* HSI clock is not divided */
#define RCC_HSI_DIV2               RCC_CR_HSIDIV_0					/* HSI clock is divided by 2 */
#define RCC_HSI_DIV4               RCC_CR_HSIDIV_1					/* HSI clock is divided by 4 */
#define RCC_HSI_DIV8               (RCC_CR_HSIDIV_1|RCC_CR_HSIDIV_0)/* HSI clock is divided by 8 */
#define RCC_HSI_DIV16              RCC_CR_HSIDIV_2					/* HSI clock is divided by 16 */
#define RCC_HSI_DIV32              (RCC_CR_HSIDIV_2|RCC_CR_HSIDIV_0)/* HSI clock is divided by 32 */
#define RCC_HSI_DIV64              (RCC_CR_HSIDIV_2|RCC_CR_HSIDIV_1)/* HSI clock is divided by 64 */
#define RCC_HSI_DIV128             (RCC_CR_HSIDIV_2|RCC_CR_HSIDIV_1|RCC_CR_HSIDIV_0) /* HSI clock is divided by 128 */

/*
 *  RCC APB Clock Source
 */
#define RCC_HCLK_DIV1              0x00000000							/* HCLK not divided */
#define RCC_HCLK_DIV2              RCC_CFGR_PPRE_2						/* HCLK divided by 2 */
#define RCC_HCLK_DIV4              (RCC_CFGR_PPRE_2 | RCC_CFGR_PPRE_0)	/* HCLK divided by 4 */
#define RCC_HCLK_DIV8              (RCC_CFGR_PPRE_2 | RCC_CFGR_PPRE_1)	/* HCLK divided by 8 */
#define RCC_HCLK_DIV16             (RCC_CFGR_PPRE_2 | RCC_CFGR_PPRE_1 | RCC_CFGR_PPRE_0) /* HCLK divided by 16 */

/*
 *  RCC AHB Clock Source
 */
#define RCC_SYSCLK_DIV1            0x00000000							/* SYSCLK not divided */
#define RCC_SYSCLK_DIV2            RCC_CFGR_HPRE_3						/* SYSCLK divided by 2 */
#define RCC_SYSCLK_DIV4            (RCC_CFGR_HPRE_3 | RCC_CFGR_HPRE_0)	/* SYSCLK divided by 4 */
#define RCC_SYSCLK_DIV8            (RCC_CFGR_HPRE_3 | RCC_CFGR_HPRE_1)	/* SYSCLK divided by 8 */
#define RCC_SYSCLK_DIV16           (RCC_CFGR_HPRE_3 | RCC_CFGR_HPRE_1 | RCC_CFGR_HPRE_0)	/* SYSCLK divided by 16 */
#define RCC_SYSCLK_DIV64           (RCC_CFGR_HPRE_3 | RCC_CFGR_HPRE_2)	/* SYSCLK divided by 64 */
#define RCC_SYSCLK_DIV128          (RCC_CFGR_HPRE_3 | RCC_CFGR_HPRE_2 | RCC_CFGR_HPRE_0)	/* SYSCLK divided by 128 */
#define RCC_SYSCLK_DIV256          (RCC_CFGR_HPRE_3 | RCC_CFGR_HPRE_2 | RCC_CFGR_HPRE_1)	/* SYSCLK divided by 256 */
#define RCC_SYSCLK_DIV512          (RCC_CFGR_HPRE_3 | RCC_CFGR_HPRE_2 | RCC_CFGR_HPRE_1 | RCC_CFGR_HPRE_0)	/* SYSCLK divided by 512 */


static const uint32_t AHBPrescTable[16] = {0U, 0U, 0U, 0U, 0U, 0U, 0U, 0U, 1U, 2U, 3U, 4U, 6U, 7U, 8U, 9U};
static const uint32_t APBPrescTable[8] =  {0U, 0U, 0U, 0U, 1U, 2U, 3U, 4U};

uint32_t SystemCoreClock = 16000000;

/*
 *  ƥ९å׻
 */
static uint32_t
RCC_GetSysClockFreq(void)
{
	uint32_t ssource = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR)) & RCC_CFGR_SWS;
	uint32_t sysclockfreq = 0;

	if(ssource == RCC_CFGR_SWS_HSI){		/* HSI source */
		uint32_t hsidiv = (sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_HSIDIV) >> 11;
		/*
		 *  HSI_DIVǳ
		 */
		sysclockfreq = HSI_VALUE / (1 << hsidiv);
	}
	else if(ssource == RCC_CFGR_SWS_HSE){	/* HSE source */
		sysclockfreq = HSE_VALUE;
	}
	else if(ssource == RCC_CFGR_SWS_PLL){	/* PLL1 source */
		uint32_t pllcfgr = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR));
		uint32_t pllsource = (pllcfgr & RCC_PLLCFGR_PLLSRC);
		uint32_t pllm = ((pllcfgr & RCC_PLLCFGR_PLLM) >> PLLM_SHIFT) + 1;
		uint32_t pllvco, pllr;

		switch (pllsource){
		case RCC_PLLSOURCE_HSE:  /* HSE used as PLL clock source */
			pllvco =  (HSE_VALUE / pllm) * ((pllcfgr & RCC_PLLCFGR_PLLN) >> PLLN_SHIFT);
			break;

		case RCC_PLLSOURCE_HSI:  /* HSI16 used as PLL clock source */
		default:                 /* HSI16 used as PLL clock source */
			pllvco = (HSI_VALUE / pllm) * ((pllcfgr & RCC_PLLCFGR_PLLN) >> PLLN_SHIFT) ;
			break;
		}
		pllr = (((pllcfgr & RCC_PLLCFGR_PLLR) >> PLLR_SHIFT) + 1);
		sysclockfreq = pllvco / pllr;
	}
	else if(ssource == RCC_CFGR_SWS_LSE){	/* LSE source */
		sysclockfreq = LSE_VALUE;
	}
	else if(ssource == RCC_CFGR_SWS_LSI){	/* LSI source */
		sysclockfreq = LSI_VALUE;
	}
	else{
		sysclockfreq = LSI_VALUE;
	}
	return sysclockfreq;
}

/*
 *  祷ƥ९å
 *  ͤϰʲ̤
 *      System Clock source            = PLL (HSI)
 *      SYSCLK(Hz)                     = 56000000
 *      HCLK(Hz)                       = 56000000
 *      AHB Prescaler                  = 1
 *      APB1 Prescaler                 = 1
 *      HSI Frequency(Hz)              = 16000000
 *      PLL_MUL                        = 4
 *      PLL_N                          = 70
 *      PLL_P                          = 10
 *      PLL_Q                          = 5
 *      PLL_R                          = 5
 *      Flash Latency(WS)              = 2
 *  return  ｪλʤ0֤
 */
static int
SetSysClock(void)
{
	uint32_t tick;
	uint32_t cfgr = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR));
	uint32_t ssource, psource, tmp;
	uint32_t FLatency = FLASH_LATENCY_2;

	/*
	 *  HSI(HSI򥷥ƥ९åȤƻѤ)
	 */
	ssource = cfgr & RCC_CFGR_SWS;
	psource = cfgr & RCC_PLLCFGR_PLLSRC;
	if(((ssource == RCC_CFGR_SWS_PLL) && (psource == RCC_PLLSOURCE_HSI)) || (ssource == RCC_CFGR_SWS_HSI)){
		/*
		 *  HSI֥졼ͤ
		 */
		sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_ICSCR), RCC_ICSCR_HSITRIM, RCC_HSICALIBRATION_DEFAULT << ICSCR_HSITRIM_SHIFT);

		/*
		 *  HSIƥ९åξ硢ǥХͤ1ˤƥƥ९å
		 */
		if(ssource == RCC_CFGR_SWS_HSI){
			sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_HSIDIV, RCC_HSI_DIV1);
			SystemCoreClock = (HSI_VALUE / (1 << (((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_HSIDIV)) >> HSIDIV_SHIFT)));
		}
	}
	else{
		/*
		 *  HSIƥ९åǤʤϡHSIνԤ
		 */
		sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_HSIDIV, RCC_HSI_DIV1);
		sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_HSION);

		/*
		 *  HSIǥԤ
		 */
		tick = 0;
		while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_HSIRDY) == 0){
			if(tick++ > HSI_TIMEOUT_VALUE)
				return -1;
			sil_dly_nse(1000);
		}

		/*
		 *  HSI֥졼ͤ
		 */
		sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_ICSCR), RCC_ICSCR_HSITRIM, RCC_HSICALIBRATION_DEFAULT << ICSCR_HSITRIM_SHIFT);
    }

	/*
	 *  PLL1
	 *  PLL1CPUåǻHSIڤؤ
	 */
	if(ssource == RCC_CFGR_SWS_PLL){
		cfgr = (cfgr & ~RCC_CFGR_SW) | RCC_SYSCLKSOURCE_HSI;
		sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), cfgr);

		/* Get Start Tick*/
		tick = 0;
		while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR)) & RCC_CFGR_SWS) != (RCC_SYSCLKSOURCE_HSI << CFGR_SWS_SHIFT)){
			if(tick++ > CLOCKSWITCH_TIMEOUT_VALUE)
				return -1;
			sil_dly_nse(1000);
		}
	}

	/*
	 *  PLL1
	 */
	sil_andw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_PLLON);

	tick = 0;
	while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_PLLRDY) != 0){
		if(tick++ > PLL_TIMEOUT_VALUE)
			return -1;
		sil_dly_nse(1000);
	}

	/*
	 *  PLL1Υե
	  */
	sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR), (RCC_PLLSOURCE_HSI | \
			((PLL_R-1)<<PLLR_SHIFT) | ((PLL_Q-1)<<24) | ((PLL_P-1)<<17) | (PLL_N << PLLN_SHIFT) | ((PLL_M-1)<<PLLM_SHIFT)));

	/*
	 *  PLL1Ƴ/PLLRå
	 */
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR), RCC_CR_PLLON);
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_PLLCFGR), RCC_PLLCFGR_PLLREN);

	/*
	 *  PLL1ǥԤ
	 */
	tick = 0;
	while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CR)) & RCC_CR_PLLRDY) == 0){
		if(tick++ > PLL_TIMEOUT_VALUE)
			return -1;
		sil_dly_nse(1000);
	}

	/*
	 *  CPU/SYSCLK/HCLK/PLL
	 */
	if(FLatency > (sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR)) & FLASH_ACR_LATENCY)){
		/*
		 *  FLASH쥤ƥ󥷥(׵ϸ⤤
		 */
		sil_modw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_LATENCY, FLatency);
		tick = 0;
		while((sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR)) & FLASH_ACR_LATENCY) != FLatency){
			if(tick++ > CLOCKSWITCH_TIMEOUT_VALUE)
				return -1;
			sil_dly_nse(1000);
		}
	}

	/*
	 *  HCLK
	 */
	cfgr  = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR)) & ~RCC_CFGR_PPRE;
	cfgr |= RCC_HCLK_DIV16;
	sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), cfgr);

	/* Set the new HCLK clock divider */
	cfgr = (cfgr & ~RCC_CFGR_HPRE) | RCC_SYSCLK_DIV1;
	sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), cfgr);

	/*
	 *  SYSCLK(PCK1)
	 */
	cfgr = (cfgr & ~RCC_CFGR_SW) | RCC_SYSCLKSOURCE_PLLCLK;
	sil_wrw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), cfgr);

	/*
	 *  SYSCLKԤ
	 */
	tick = 0;
	while((sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR)) & RCC_CFGR_SWS) != (RCC_SYSCLKSOURCE_PLLCLK << CFGR_SWS_SHIFT)){
		if(tick++ > CLOCKSWITCH_TIMEOUT_VALUE)
			return -1;
		sil_dly_nse(1000);
	}

	if(FLatency < (sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR)) & FLASH_ACR_LATENCY)){
		/*
		 *  FLASH쥤ƥ󥷥(׵ϸ㤤
		 */
		sil_modw_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR), FLASH_ACR_LATENCY, FLatency);
		while((sil_rew_mem((uint32_t *)(TADR_FLASH_R_BASE+TOFF_FLASH_ACR)) & FLASH_ACR_LATENCY) != FLatency){
			if(tick++ > CLOCKSWITCH_TIMEOUT_VALUE)
				return -1;
			sil_dly_nse(1000);
		}
	}

	/*
	 *  PCLK1 HCLKʬ
	 */
	sil_modw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR), RCC_CFGR_PPRE, RCC_HCLK_DIV1);

	/*
	 *  ƥ९åƷ׻
	 */
	cfgr  = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR));
	SystemCoreClock = (RCC_GetSysClockFreq() >> ((AHBPrescTable[(cfgr & RCC_CFGR_HPRE) >> CFGR_HPRE_SHIFT]) & 0x1F));
	(void)(tmp);
	return 0;
}

/*
 *  ٥ΥåȰ¸ν
 *
 *  ȥåץ⥸塼ǡν˸ƤӽФ롥
 */
void
SystemInit(void)
{
	uint32_t tmp;
	/* Configure the Vector Table location add offset address ------------------*/
	sil_wrw_mem((uint32_t *)(TADR_SCB_BASE+TOFF_SCB_VTOR), (FLASH_BASE));

	SetSysClock();

	/*
	 *  SYSCFGPWRΥåͭ
	 */
	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APBENR2), RCC_APBENR2_SYSCFGEN);
	tmp = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APBENR2));

	sil_orw_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APBENR1), RCC_APBENR1_PWREN);
	tmp = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_APBENR1));
	(void)(tmp);
}

/*
 *  CortexM0+LGϥ٥ѹǽ
 */
void SetVector(void *addr)
{
	sil_wrw_mem((uint32_t *)(TADR_SCB_BASE+TOFF_SCB_VTOR), (uint32_t)addr);
}

/*
 *  ߤSYSCLK, HCLK, PCLK1, PCLK2ͤФؿ
 */
void
RCC_GetClocksFreq(RCC_ClocksTypeDef* RCC_Clocks)
{
	uint32_t cfgr = sil_rew_mem((uint32_t *)(TADR_RCC_BASE+TOFF_RCC_CFGR));

	RCC_Clocks->SYSCLK_Frequency = RCC_GetSysClockFreq();
	RCC_Clocks->HCLK_Frequency  = SystemCoreClock;
	RCC_Clocks->PCLK1_Frequency = SystemCoreClock >> (APBPrescTable[(cfgr & RCC_CFGR_PPRE) >> 12]);
}

